import numpy as np
import pandas as pd
import os
from sklearn.model_selection import train_test_split
from ampligraph.latent_features import DistMult, ComplEx, HolE, save_model, restore_model
from ampligraph.evaluation import train_test_split_no_unseen, evaluate_performance, mrr_score, hits_at_n_score, select_best_model_ranking
import pyiwn
from itertools import combinations
import pickle5 as pickle
import networkx as nx
from scipy.special import expit
import collections

def get_rel(i):
    rel_dict = {1: 'ability_verb.noun',
                2: 'also_see.adjective',
                3: 'also_see.adverb',
                4: 'also_see.verb',
                5: 'attributes.noun',
                6: 'capability_verb.noun',
                7: 'causative.verb',
                8: 'entailment.verb',
                9: 'function_verb.noun',
                10: 'holo_component_object.noun',
                11: 'holo_feature_activity.noun',
                12: 'holo_member_collection.noun',
                13: 'holo_place_area.noun',
                14: 'holo_portion_mass.noun',
                15: 'holo_position_area.noun',
                16: 'holo_stuff_object.noun',
                17: 'hypernymy.noun',
                18: 'hypernymy.verb',
                19: 'hyponymy.noun',
                20: 'mero_component_object.noun',
                21: 'mero_feature_activity.noun',
                22: 'mero_member_collection.noun',
                23: 'mero_place_area.noun',
                24: 'mero_portion_mass.noun',
                25: 'mero_position_area.noun',
                26: 'mero_stuff_object.noun',
                27: 'modifies_noun.adjective',
                28: 'modifies_verb.adjverb',
                29: 'similar.adjective',
                30: 'troponymy.verb',
                31: 'synonymy'}
    return rel_dict[i]

def generate_triples(tuple_list, rel_id):
    triple_list = []
    for tup in tuple_list:
        triple = (tup[0], rel_id, tup[1])
        triple_list.append(triple)
    return triple_list

def generate_allcombs(list1, list2):
    res = []
    for val1 in list1:
        for val2 in list2:
            tup = (val1, val2)
            res.append(tup)
    return res

def convert_iwn():
   beng_iwn = pyiwn.IndoWordNet(lang = pyiwn.Language.BENGALI)
   all_syns = beng_iwn.all_synsets()
   edges = pd.read_csv("../data/BengIwnData/edges.csv", header = None, delimiter = '\t', names = ['ent1','ent2', 'rel'])
   X = pd.DataFrame()
   X['ent1'] = edges['ent1']
   X['rel'] = edges['rel']
   X['ent2'] = edges['ent2']
   print("The dataframe X: \n", X.head())
   X = X.astype(str)
   print("Changing the edge structure from (synset -> synset) to (word -> word)")

   # converting synsets list into a hashmap
   syn_dict = dict()
   for val in all_syns:
     syn_dict[val.synset_id()] = val
     

   words_set = set()
   word_edges = set()
   visited_syns = set()
   X = X.astype(int)
   for i, row in X.iterrows():
       syn1_id = row['ent1']
       syn1 = syn_dict[int(syn1_id)]
       word_list1 = syn1.lemma_names()
       rel = row['rel']
       syn2_id = row['ent2']
       syn2 = syn_dict[int(syn2_id)]
       word_list2 = syn2.lemma_names()

       if(syn1_id not in visited_syns):
           # generating tuples for synonymy relation - id = 31
           synonymy_combs1 = combinations(word_list1, 2)
           synonymy_triples1 = generate_triples(synonymy_combs1, '31') # add to set
           for i, val in enumerate(synonymy_triples1):
               word_edges.add(val)
               visited_syns.add(syn1_id)

       if(syn2_id not in visited_syns):
           synonymy_combs2 = combinations(word_list2, 2)
           synonymy_triples2 = generate_triples(synonymy_combs2, '31') # add to set
           for i, val in enumerate(synonymy_triples2):
               word_edges.add(val)
               visited_syns.add(syn2_id)

  # enumerating the same relation for all words in both synset lists
       word_combinations = generate_allcombs(word_list1, word_list2)
       word_triples = generate_triples(word_combinations, str(rel)) # add to set

       for i, val in enumerate(word_triples):
           word_edges.add(val)
    
   return word_edges

positive_filter = []
dt_train = []
dt_test = []
reduced_dt_test = []
count_unseen = 0
def label_DT():
    #loading subgraph
    count_unseen = 0
    count_seen = 0
    filen = open("./subgraph3.pkl", "rb")
    subgraph = pickle.load(filen)

    iwn = convert_iwn()
    iwn_arr = np.asarray(list(iwn))
    iwn_dict = collections.defaultdict(int)

    for i,row in enumerate(iwn_arr):
        key = tuple((row[0], row[2]))
        iwn_dict[key] = int(row[1])

    rel_range = set()
    ent_range = set()
    for u,v in subgraph.edges():
        cur = tuple((u,v))
        if iwn_dict[cur] == 0:
            # unlabelled edge
            # assign all edge labels and add to test set
            count_unseen+=1
            for rel_id in range(1,32):
                label = get_rel(rel_id)
                triple = tuple((u, label, v))
                dt_test.append(triple)
        else:
            #labelled edge
            count_seen+=1
            ent_range.add(u)
            ent_range.add(v)
            label = get_rel(iwn_dict[cur])
            rel_range.add(label)
            triple = tuple((u, label, v))
            dt_train.append(triple)
            #print(f"Number of unique relations from WordNet, mapped to DT = {len(rel_range)}")

    for i,val in enumerate(dt_test):
        ent1 = val[0]
        ent2 = val[2]
        rel = val[1]
        if ent1 in ent_range and ent2 in ent_range and rel in rel_range:
            reduced_dt_test.append(val)

    print(f"Original test size = {len(dt_test)}")
    print(f"Reduced test size = {len(reduced_dt_test)}")
    print(f"Train size = {len(dt_train)}")
    print(f"Count of unseen edges = {count_unseen}")
    print(f"Count of seen edges = {count_seen}")

def evaluate(model, X_test, positive_filter) :
  batchsize=100000
  start = 0
  mrr_test = []
  hits_at_1 = []
  hits_at_3 = []
  hits_at_10 = []
  for i in range(0, (len(X_test)//batchsize)+1):
      end = start+batchsize
      if end > len(X_test):
          end = len(X_test)
      cur_triples = X_test[start:end]

      test_ranks = evaluate_performance(cur_triples, model=model, filter_triples = positive_filter, filter_unseen= True)
      mrr_test.append(mrr_score(test_ranks))
      hits_at_1.append(hits_at_n_score(test_ranks, 1))
      hits_at_3.append(hits_at_n_score(test_ranks, 3))
      hits_at_10.append(hits_at_n_score(test_ranks, 10))
      start+=batchsize
      print(f"####Completed {i}th iteration!####\n")
    
  print("Results:")
  print(f"Test mrr = {np.mean(np.asarray(mrr_test))}")
  print(f"Test hits_at_1 score = {np.mean(np.asarray(hits_at_1))}")
  print(f"Test hits_at_3 score = {np.mean(np.asarray(hits_at_3))}")
  print(f"Test hits_at_10 score = {np.mean(np.asarray(hits_at_10))}")
 
def get_unseen_scores(model, unseen_triples):
    batch_size = 50000
    start = 0
    final_scores = pd.DataFrame(columns = ['triple', 'score', 'prob'])
    for i in range(0, (len(unseen_triples)//batch_size) +1):
        end = start+batch_size
        if end > len(unseen_triples):
            end = len(unseen_triples)
        cur_triples = unseen_triples[start:end]
        #ranks_unseen = evaluate_performance(
        #        unseen_triples,
        #        model = model,
        #        filter_triples = unseen_filter,
        #        filter_unseen=True
        #        )   
        scores = model.predict(cur_triples)
        # converting to probabilities
        probs = expit(scores)
        score_df = pd.DataFrame(list(zip([' '.join(x) for x in cur_triples],
            np.squeeze(scores),
            np.squeeze(probs))),
            columns = ['triple', 'score', 'prob']).sort_values("score", ascending = False)
        final_scores = final_scores.append(score_df)
        print(f"Stacked final scores df shape = {final_scores.shape}")
        start+=batch_size
        print(f"###Iteration {i} completed!###")
    sorted_scores = final_scores.sort_values(by=["score"], ascending = False)
    del final_scores
    print(sorted_scores.head())
    sorted_scores.to_csv("./dt2_dist.csv")
    return sorted_scores

label_DT()
#test_size = (0.2)*len(dt)
train = np.asarray(dt_train)
test = np.asarray(reduced_dt_test)
positive_filter = train
#train, test = train_test_split_no_unseen(dt, seed = 1, test_size = test_size)
unseen_filter = np.array(list({tuple(i) for i in np.vstack((positive_filter, test))}))

dist_model = DistMult(batches_count= 100, epochs=10, k= 100, eta= 50, optimizer= 'adagrad', loss= 'nll', regularizer= 'LP', verbose= False, optimizer_params= {'lr': 0.1}, regularizer_params= {'p': 2, 'lambda': 0.001}, embedding_model_params= {'negative_corruption_entities': 'all'})

dist_model.fit(train)
evaluate(dist_model, test, positive_filter)
#scores = get_unseen_scores(dist_model, test)






























































    

